
/*
 *  finddup is a simple utility to find duplicated files, files common
 *  to several directories, or files present in one directory and not
 *  in another one.
 *
 *  Copyright (c) 2010, 2011 Francois Fleuret
 *  Written by Francois Fleuret <francois@fleuret.org>
 *
 *  This file is part of finddup.
 *
 *  finddup is free software: you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License version 3 as
 *  published by the Free Software Foundation.
 *
 *  finddup is distributed in the hope that it will be useful, but WITHOUT
 *  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 *  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
 *  License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with finddup.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#define VERSION_NUMBER "1.2"

#define _BSD_SOURCE

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/wait.h>
#include <dirent.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <sys/ioctl.h>
#include <locale.h>
#include <getopt.h>
#include <fcntl.h>

/* 1M really helps compared to 64k */
#define READ_BUFFER_SIZE (1024 * 1024)

#define PROGRESS_BUFFER_SIZE 1024

typedef int64_t size_sum_t;

/* Yeah, global variables! */

int ignore_dotfiles = 0; /* 1 means ignore files and directories
                            starting with a dot */

int ignore_empty_files = 0; /* 1 means ignore empty files */

int show_realpaths = 0; /* 1 means ignore files and directories
                            starting with a dot */

int show_progress = 0; /* 1 means show a progress bar when we are in a
                          tty */

int show_hits = 1; /* 1 means to show the files from dir2
                      corresponding to the ones from dir1 */

int show_groups = 1; /* 1 means to show the group IDs when printing
                        file names */

int same_inodes_are_different = 0; /* 1 means that comparison between
                                      two files with same inode will
                                      always be false */

int sort_by_time = 0; /* 1 means to sort files in each group according
                         to the modification time */

int trim_first = 0; /* remove the first entry in each group */

char *command_to_exec = 0; /* the name of the command to exec for each
                              group of identical files */

char *result_file_prefix = 0; /* The prefix to use to write result */

/********************************************************************/

/* malloc with error checking.  */

void *safe_malloc(size_t n) {
  void *p = malloc(n);
  if (!p && n != 0) {
    fprintf(stderr,
            "finddup: Can not allocate memory: %s\n", strerror(errno));
    exit(EXIT_FAILURE);
  }
  return p;
}

/********************************************************************/

int ignore_entry(const char *name) {
  return
    strcmp(name, ".") == 0 ||
    strcmp(name, "..") == 0 ||
    (ignore_dotfiles && name[0] == '.' && name[1] != '/');
}

/**********************************************************************/

struct file_node {
  struct file_node *next;
  char *name;
  size_t size;
  time_t atime, mtime, ctime;
  ino_t inode;
  int group_id; /* one per identical file content */
  int dir_id; /* 1 for DIR1, and 2 for DIR2 */
};

void file_list_delete(struct file_node *head) {
  struct file_node *next;
  while(head) {
    next = head->next;
    free(head->name);
    free(head);
    head = next;
  }
}

int file_list_length(struct file_node *head) {
  int l = 0;
  while(head) {
    l++;
    head = head->next;
  }
  return l;
}

/**********************************************************************/

int same_content(struct file_node *f1, struct file_node *f2,
                 char *buffer1, char *buffer2) {
  int fd1, fd2, s1, s2;

  fd1 = open(f1->name, O_RDONLY);
  fd2 = open(f2->name, O_RDONLY);

  if(fd1 >= 0 && fd2 >= 0) {
    while(1) {
      s1 = read(fd1, buffer1, READ_BUFFER_SIZE);
      s2 = read(fd2, buffer2, READ_BUFFER_SIZE);

      if(s1 < 0 || s2 < 0) {
        close(fd1);
        close(fd2);
        return 0;
      }

      if(s1 == s2) {
        if(s1 == 0) {
          close(fd1);
          close(fd2);
          return 1;
        } else {
          if(memcmp(buffer1, buffer2, s1)) {
            /* printf("size_to_read = %d\n", size_to_read); */
            close(fd1);
            close(fd2);
            return 0;
          }
        }
      } else {
        fprintf(stderr,
                "finddup: Different read size without error on files of same size.\n");
        exit(EXIT_FAILURE);
      }
    }
  } else {

    if(fd1 < 0) {
      fprintf(stderr,
              "finddup: Can not open \"%s\" error: %s\n",
              f1->name,
              strerror(errno));
    }

    if(fd2 < 0) {
      fprintf(stderr,
              "finddup: Can not open \"%s\" error: %s\n",
              f2->name,
              strerror(errno));
    }

    exit(EXIT_FAILURE);
  }
}

int same_files(struct file_node *f1, struct file_node *f2,
               char *buffer1, char *buffer2) {
  if(same_inodes_are_different && f1->inode == f2->inode) {
    return 0;
  }

  return f1->size == f2->size && same_content(f1, f2, buffer1, buffer2);
}

/**********************************************************************/

struct file_node *scan_directory_rec(struct file_node *tail, const char *name) {
  DIR *dir;
  struct dirent *dir_e;
  struct stat sb;
  struct file_node *tmp;
  char subname[PATH_MAX + 1];

  if(lstat(name, &sb) != 0) {
    fprintf(stderr, "finddup: Can not stat \"%s\": %s\n", name, strerror(errno));
    exit(EXIT_FAILURE);
  }

  if(S_ISLNK(sb.st_mode)) {
    return tail;
  }

  dir = opendir(name);

  if(dir) {
    while((dir_e = readdir(dir))) {
      if(!ignore_entry(dir_e->d_name)) {
        snprintf(subname, PATH_MAX, "%s/%s", name, dir_e->d_name);
        tail = scan_directory_rec(tail, subname);
      }
    }
    closedir(dir);
  } else {
    if(S_ISREG(sb.st_mode)) {
      if(!ignore_entry(name)) {
        if(!ignore_empty_files || sb.st_size > 0) {
          tmp = safe_malloc(sizeof(struct file_node));
          tmp->next = tail;
          tmp->name = strdup(name);
          tmp->size = sb.st_size;
          tmp->atime = sb.st_atime;
          tmp->mtime = sb.st_mtime;
          tmp->ctime = sb.st_ctime;
          tmp->inode = sb.st_ino;
          tmp->group_id = -1;
          tmp->dir_id = -1;
          tail = tmp;
        }
      }
    }
  }

  return tail;
}

struct file_node *scan_directory(struct file_node *tail, const char *name) {
  struct file_node *result;
  int length;

  if(show_progress) {
    fprintf(stderr, "Scanning '%s' ... ", name);
  }

  result = scan_directory_rec(tail, name);
  length = file_list_length(result);

  if(show_progress) {
    fprintf(stderr, "done (%d file%s).\n",
            length, (length > 1 ? "s" : ""));
  }


  return result;
}

void write_one_entry_to_file(FILE *out, struct file_node *node) {
  char tmp[PATH_MAX + 1];
  if(show_realpaths) {
    if(realpath(node->name, tmp)) {
      if(show_groups) {
        fprintf(out, "%d %s\n", node->group_id, tmp);
      } else {
        fprintf(out, "%s\n", tmp);
      }
    } else {
      fprintf(stderr,
              "finddup: Can not get the realpath of \"%s\": %s\n",
              node->name,
              strerror(errno));
      exit(EXIT_FAILURE);
    }
  } else {
    if(show_groups) {
      fprintf(out, "%d %s\n", node->group_id, node->name);
    } else {
      fprintf(out, "%s\n", node->name);
    }
  }
}

int compare_nodes(const void *x1, const void *x2) {
  const struct file_node **f1, **f2;

  f1 = (const struct file_node **) x1;
  f2 = (const struct file_node **) x2;

  if((*f1)->group_id < (*f2)->group_id) {
    return -1;
  } else if((*f1)->group_id > (*f2)->group_id) {
    return 1;
  } else {
    if(sort_by_time) {
      if((*f1)->mtime < (*f2)->mtime) {
        return -1;
      } else if((*f1)->mtime > (*f2)->mtime) {
        return 1;
      } else {
        return 0;
      }
    } else {
      if((*f1)->dir_id < (*f2)->dir_id) {
        return -1;
      } else if((*f1)->dir_id > (*f2)->dir_id) {
        return 1;
      } else {
        return 0;
      }
    }
  }
}

void exec_command(int nb, struct file_node **nodes) {
  char **args;
  int max_group_size = 0, group_size = 0, m, n, status;
  pid_t pid;

  for(n = 0; n < nb; n++) {
    if(nodes[n]->group_id != nodes[n-1]->group_id) {
      group_size = 0;
    }
    group_size++;
    if(group_size > max_group_size) {
      max_group_size = group_size;
    }
  }

  args = safe_malloc((max_group_size + 2) * sizeof(char *));
  args[0] = command_to_exec;

  n = 0;
  while(n < nb) {
    m = n;
    if(trim_first) { m++; }
    while(n < nb && nodes[n]->group_id == nodes[m]->group_id) {
      if(n >= m) {
        args[n - m + 1] = nodes[n]->name;
      }
      n++;
    }
    args[n - m + 1] = 0;
    pid = fork();
    if(pid < 0) {
    } else if(pid == 0) {
      if(execvp(command_to_exec, args) < 0) {
        perror("execvp");
        exit(EXIT_FAILURE);
      }
    } else {
      while(wait(&status) != pid);
      if(status > 0) {
        exit(EXIT_FAILURE);
      }
    }
  }

  free(args);
}

void write_groups_in_files(int nb, struct file_node **nodes) {
  FILE *file = 0;
  int current_group = -1, n, first_of_group;
  char filename[PATH_MAX + 1];

  for(n = 0; n < nb; n++) {
    first_of_group = (n == 0);
    if(nodes[n]->group_id != current_group) {
      if(file) { fclose(file); }
      sprintf(filename, "%s%06d", result_file_prefix, nodes[n]->group_id);
      file = fopen(filename, "w");
      current_group = nodes[n]->group_id;
      printf("Writing %s.\n" , filename);
      first_of_group = 1;
    }

    if(!trim_first || !first_of_group) {
      write_one_entry_to_file(file, nodes[n]);
    }
  }

  if(file) { fclose(file); }
}

void print_result(struct file_node *list1, struct file_node *list2) {
  struct file_node *node1, *node2;
  struct file_node **nodes;
  int nb, n, first_of_group;

  nb = 0;
  for(node1 = list1; node1; node1 = node1->next) {
    if(node1->group_id >= 0) { nb++; }
  }

  if(list2) {
    for(node2 = list2; node2; node2 = node2->next) {
      if(node2->group_id >= 0) { nb++; }
    }
  }

  nodes = safe_malloc(nb * sizeof(struct file_node *));

  n = 0;
  for(node1 = list1; node1; node1 = node1->next) {
    if(node1->group_id >= 0) {
      nodes[n++] = node1;
    }
  }

  if(list2) {
    for(node2 = list2; node2; node2 = node2->next) {
      if(node2->group_id >= 0) {
        nodes[n++] = node2;
      }
    }
  }

  qsort(nodes, nb, sizeof(struct file_node *), compare_nodes);

  if(command_to_exec) {
    exec_command(nb, nodes);
  } else if(result_file_prefix) {
    write_groups_in_files(nb, nodes);
  } else {
    for(n = 0; n < nb; n++) {
      first_of_group = (n == 0);
      if(n > 0 && nodes[n]->group_id != nodes[n-1]->group_id) {
        if(!show_groups) {
          printf("\n");
        }
        first_of_group = 1;
      }
      if(!trim_first || !first_of_group) {
        write_one_entry_to_file(stdout, nodes[n]);
      }
    }
  }

  free(nodes);
}

struct progress_state {
  int bar_width;
  int nb_values, value;
  int last_position;
};

void print_progress(struct progress_state *state) {
  int position, k, normalizer;
  struct winsize win;
  char buffer[PROGRESS_BUFFER_SIZE];
  char *s;

  normalizer = (state->nb_values > 1 ? state->nb_values - 1 : 1);

  if(show_progress) {
    /* We use the previous bar_width to compute the position, so that
       we avoid doing too many ioctls */
    position = (state->bar_width * state->value) / normalizer;
    if(state->bar_width <= 0 || position != state->last_position) {
      if(!ioctl (STDERR_FILENO, TIOCGWINSZ, (char *) &win)) {
        /* Something weird is going on if the previous test is wrong */
        if(win.ws_col >= PROGRESS_BUFFER_SIZE - 3) {
          state->bar_width = PROGRESS_BUFFER_SIZE - 10;
        } else {
          state->bar_width = win.ws_col - 7;
        }
        position = (state->bar_width * state->value) / normalizer;
        state->last_position = position;
        s = buffer;
        for(k = 0; k < position; k++) {
          *(s++) = '+';
        }
        for(; k < state->bar_width; k++) {
          *(s++) = '-';
        }

        /* We need four % because of the fprintf that follows */
        sprintf(s, " [%3d%%]\r",
                (100 * state->value) / normalizer);

        fprintf(stderr, "%s", buffer);
      }
    }
  }
}

void start(const char *dirname1, const char *dirname2) {
  struct file_node *list1, *list2;
  struct file_node *node1, *node2;
  struct progress_state progress_state;
  int not_in, found;
  int nb_groups, nb_nodes;
  int list1_length;

  char *buffer1 = safe_malloc(sizeof(char) * READ_BUFFER_SIZE);
  char *buffer2 = safe_malloc(sizeof(char) * READ_BUFFER_SIZE);

  not_in = 0;

  list1 = scan_directory(0, dirname1);
  list1_length = file_list_length(list1);

  if(dirname2) {
    if(strncmp(dirname2, "not:", 4) == 0) {
      not_in = 1;
      /* groups are not computed in the not: mode */
      show_groups = 0;
      dirname2 += 4;
    } else if(strncmp(dirname2, "and:", 4) == 0) {
      dirname2 += 4;
    }
    list2 = scan_directory(0, dirname2);
  } else {
    list2 = list1;
  }

  if(show_progress) {
    fprintf(stderr,
            "Now looking for identical files (this may take a while).\n");
  }

  nb_groups = 0;
  nb_nodes = 0;

  progress_state.bar_width = -1;
  progress_state.last_position = -1;
  progress_state.nb_values = list1_length;

  if(not_in) {
    for(node1 = list1; node1; node1 = node1->next) {
      progress_state.value = nb_nodes;
      print_progress(&progress_state);
      nb_nodes++;

      found = 0;

      for(node2 = list2; !found && node2; node2 = node2->next) {
        if(same_files(node1, node2, buffer1, buffer2)) {
          found = 1;
        }
      }

      if(!found) {
        if(show_realpaths) {
          printf("%s\n", realpath(node1->name, 0));
        } else {
          printf("%s\n", node1->name);
        }
      }
    }

  } else {
    for(node1 = list1; node1; node1 = node1->next) {
      progress_state.value = nb_nodes;
      print_progress(&progress_state);
      nb_nodes++;

      for(node2 = list2; node2; node2 = node2->next) {
        if(node1->group_id < 0 || node2->group_id < 0) {
          if(same_files(node1, node2, buffer1, buffer2)) {
            if(node1->group_id < 0) {
              if(node2->group_id >= 0) {
                node1->group_id = node2->group_id;
              } else {
                node1->group_id = nb_groups;
                node1->dir_id = 1;
                nb_groups++;
              }
            }
            if(node2->group_id < 0) {
              node2->group_id = node1->group_id;
              node2->dir_id = 2;
            }
          }
        }
      }
    }
  }

  if(show_progress) {
    fprintf(stderr, "\n");
  }

  if(dirname2) {
    print_result(list1, list2);
    file_list_delete(list1);
    file_list_delete(list2);
  } else {
    print_result(list1, 0);
    file_list_delete(list1);
  }

  free(buffer1);
  free(buffer2);
}

void usage(FILE *out) {
  fprintf(out, "Usage: finddup [OPTION]... [DIR1 [[and:|not:]DIR2]]\n");
  fprintf(out, "Version %s (%s)\n", VERSION_NUMBER, UNAME);
  fprintf(out, "Without DIR2, lists duplicated files found in DIR1, or the current directory if DIR1 is not provided. With DIR2, lists files common to both directories. With the not: prefix, lists files found in DIR1 which do not exist in DIR2. The and: prefix is the default and should be used only if you have a directory starting with 'not:'\n");
  fprintf(out, "\n");
  /*            01234567890123456789012345678901234567890123456789012345678901234567890123456789*/
  fprintf(out, "   -v, --version              prints the version number and exit\n");
  fprintf(out, "   -h, --help                 show this help\n");
  fprintf(out, "   -d, --ignore-dots          ignore dot files and directories\n");
  fprintf(out, "   -0, --ignore-empty         ignore empty files\n");
  fprintf(out, "   -c, --hide-matchings       do not show which files in DIR2 corresponds to\n");
  fprintf(out, "                              those in DIR1\n");
  fprintf(out, "   -g, --no-group-ids         do not show the file groups\n");
  fprintf(out, "   -t, --time-sort            sort according to modification time in each group\n");
  fprintf(out, "   -q, --trim-first           do not show the first file in each group\n");
  fprintf(out, "   -p, --show-progress        show progress\n");
  fprintf(out, "   -r, --real-paths           show the real file paths\n");
  fprintf(out, "   -i, --same-inodes-are-different\n");
  fprintf(out, "                              consider files with same inode as different\n");
  fprintf(out, "   -e <command>, --exec <command>\n");
  fprintf(out, "                              execute the provided command for each group of\n");
  fprintf(out, "                              identical files, with their names as arguments\n");
  fprintf(out, "   -f <string>, --result-prefix <string>\n");
  fprintf(out, "                              for each group of identical files, write one\n");
  fprintf(out, "                              result file whose name is the given prefix string\n");
  fprintf(out, "                              followed by the group number, and containing\n");
  fprintf(out, "                              one filename per line\n");
  fprintf(out, "\n");
  fprintf(out, "Report bugs and comments to <francois@fleuret.org>.\n");
}

/**********************************************************************/

static struct option long_options[] = {
  { "version", no_argument, 0, 'v' },
  { "help", no_argument, 0, 'h' },
  { "same-inodes-are-different", no_argument, 0, 'i' },
  { "real-paths", no_argument, 0, 'r' },
  { "hide-matchings", no_argument, 0, 'c' },
  { "no-group-ids", no_argument, 0, 'g' },
  { "time-sort", no_argument, 0, 't' },
  { "trim-first", no_argument, 0, 'q' },
  { "ignore-dots", no_argument, 0, 'd' },
  { "ignore-empty", no_argument, 0, '0' },
  { "show-progress", no_argument, 0, 'p' },
  { "exec", 1, 0, 'e' },
  { "result-prefix", 1, 0, 'f' },
  { 0, 0, 0, 0 }
};

int main(int argc, char **argv) {
  int c;

  setlocale (LC_ALL, "");

  while ((c = getopt_long(argc, argv, "vhircgtqd0pme:f:",
                          long_options, NULL)) != -1) {
    switch (c) {

    case 'v':
      printf("finddup version %s (%s)\n", VERSION_NUMBER, UNAME);
      exit(EXIT_SUCCESS);
      break;

    case 'h':
      usage(stdout);
      exit(EXIT_SUCCESS);

      break;

    case 'd':
      ignore_dotfiles = 1;
      break;

    case '0':
      ignore_empty_files = 1;
      break;

    case 'r':
      show_realpaths = 1;
      break;

    case 'i':
      same_inodes_are_different = 1;
      break;

    case 'g':
      show_groups = 0;
      break;

    case 't':
      sort_by_time = 1;
      break;

    case 'q':
      trim_first = 1;
      break;

    case 'p':
      show_progress = 1;
      break;

    case 'c':
      show_hits = 0;
      break;

    case 'e':
      if(command_to_exec != 0) {
        free(command_to_exec);
      }
      command_to_exec = strdup(optarg);
      break;

    case 'f':
      if(result_file_prefix != 0) {
        free(result_file_prefix);
      }
      result_file_prefix = strdup(optarg);
      show_groups = 0;
      break;

    default:
      usage(stderr);
      exit(EXIT_FAILURE);
    }
  }

  if(!isatty(STDERR_FILENO)) {
    show_progress = 0;
  }

  if(optind + 2 == argc) {
    start(argv[optind], argv[optind + 1]);
  } else if(optind + 1 == argc) {
    same_inodes_are_different = 1;
    start(argv[optind], 0);
  } else if(optind == argc) {
    same_inodes_are_different = 1;
    start(".", 0);
  } else {
    usage(stderr);
    exit(EXIT_FAILURE);
  }

  exit(EXIT_SUCCESS);
}
